#include<stdio.h>
#include<string.h>
#include<stdlib.h>

#include "types.h"
#include "common.h"

#include "raster.h"


unsigned NearAvailPlanes(unsigned n)
{
  if(n<1) return(n);
  if(n<=1) return(1);
  if(n<=2) return(2);
  if(n<=4) return(4);
  if(n<=8) return(8);
  if(n<=16) return(16);
  if(n<=24) return(24);
  if(n<=32) return(32);
  return(n);
}


void AbstractError(void)
{
 printf("Abstract function called!\n");abort();
}


void Raster1DAbstract::Erase1DStub(void)
{
 if(Data1D!=NULL) free(Data1D);
 Data1D=NULL;
}


void Raster1DAbstract::Cleanup(void)
{
 if(Data1D==NULL) return;
 memset(Data1D,0,Size1D*GetPlanes()/8);
}


void Raster1DAbstract::Get(Raster1DAbstract &R1)
{
int i,maxi;
maxi=Size1D>R1.Size1D?Size1D:R1.Size1D;
for(i=0;i<maxi;i++)
	{
	R1.SetValue1D(i,GetValue1D(i));
	}
}

void Raster1DAbstract::Set(Raster1DAbstract &R1)
{
int i,maxi;
int shift;
  maxi=Size1D>R1.Size1D?Size1D:R1.Size1D;
  shift = GetPlanes() - R1.GetPlanes();
  for(i=0;i<maxi;i++)
    {
    if(shift>0)
      SetValue1D(i,R1.GetValue1D(i)<<shift);
    else
      SetValue1D(i,R1.GetValue1D(i)>>(-shift));
    }
}


void Raster1DAbstract::Set16Bit(void *Buffer16Bit)
{
WORD *N_Buffer16Bit = (WORD *)Buffer16Bit;
int i;
for(i=0;i<Size1D;i++)
	{
	SetValue1D(i,*N_Buffer16Bit++);
	}
}


void Raster1DAbstract::Get8Bit(void *Buffer8Bit)
{
BYTE *N_Buffer8Bit = (BYTE *)Buffer8Bit;
int i;
for(i=0;i<Size1D;i++)
	{
	*N_Buffer8Bit++ = GetValue1D(i);
	}
}

void Raster1DAbstract::Get16Bit(void *Buffer16Bit)
{
WORD *N_Buffer16Bit = (WORD *)Buffer16Bit;
int i;
int shift;

  shift = 16-GetPlanes();

  if(shift>0)
    for(i=0;i<Size1D;i++)
      {
      *N_Buffer16Bit++ = (WORD)(GetValue1D(i) << shift);
      }
  else
    {
    shift = -shift;
    for(i=0;i<Size1D;i++)
      {
      *N_Buffer16Bit++ = (WORD)(GetValue1D(i) >> shift);
      }
    }  
}


void Raster1DAbstract::Get24Bit(void *Buffer24Bit)
{
BYTE *N_Buffer24Bit = (BYTE *)Buffer24Bit;
DWORD val;
int i;
int shift;

  if(Buffer24Bit==NULL) return;
  shift = 24-GetPlanes();

  for(i=0;i<Size1D;i++)
	{
        val = GetValue1D(i);
        if(shift>0) val <<= shift;
               else val >>= -shift;
	*N_Buffer24Bit++ = val&0xFF;
        *N_Buffer24Bit++ = (val>>8)&0xFF;
        *N_Buffer24Bit++ = (val>>16)&0xFF;
	}
}


void Raster1DAbstract::Get32Bit(void *Buffer32Bit)
{
DWORD *N_Buffer32Bit = (DWORD *)Buffer32Bit;
int i;
for(i=0;i<Size1D;i++)
	{
	*N_Buffer32Bit++ = GetValue1D(i);
	}
}


/* Specialised Raster 1D modules */
/* 1 bit plane */
class Raster1D_1Bit:virtual public Raster1DAbstract
	{
public: virtual void Allocate1D(int NewSize1D)	{Data1D=malloc((NewSize1D+7)/8);};

	virtual int GetPlanes(void) const {return(1);};

	virtual unsigned long GetValue1D(int x) const;
	virtual void SetValue1D(int x, unsigned long NewValue);
	};

void Raster1DAbstract::Get1Bit(void *Buffer1Bit)
{
 Raster1D_1Bit R1;
 R1.Data1D=Buffer1Bit;R1.Size1D=Size1D;
 Get(R1);
 R1.Data1D=NULL;
}

void Raster1DAbstract::Peel1Bit(void *Buffer1Bit,char plane)
{
Raster1D_1Bit R1;
int i,maxi;

 R1.Data1D=Buffer1Bit;R1.Size1D=Size1D;
 maxi=Size1D>R1.Size1D?Size1D:R1.Size1D;
 for(i=0;i<maxi;i++)
	{
	R1.SetValue1D(i,(GetValue1D(i)>>plane)&1);
	}
 R1.Data1D=NULL;
}

void Raster1DAbstract::Join1Bit(void *Buffer1Bit,char plane)
{
 Raster1D_1Bit R1;
 int i,maxi;
 long mask;

 R1.Data1D=Buffer1Bit;R1.Size1D=Size1D;
 maxi=Size1D>R1.Size1D?Size1D:R1.Size1D;

 mask=~(1l<<plane);
 for(i=0;i<maxi;i++)
	{
	SetValue1D(i,(GetValue1D(i) & mask) | (R1.GetValue1D(i) << plane));
	}
 R1.Data1D=NULL;
}


void Raster1D_1Bit::SetValue1D(int x, unsigned long NewValue)
{
BYTE *b=(BYTE *)Data1D;
 b+= x >> 3;
 if(NewValue==0) *b = *b & ~(0x80 >>(x & 0x07));
	    else *b = *b |  (0x80 >>(x & 0x07));
}


unsigned long Raster1D_1Bit::GetValue1D(int x) const
{
BYTE *b=(BYTE *)Data1D;

 if(x>=Size1D || Data1D==NULL) return(0);
 b+= x >> 3;
 if((*b & (0x80 >>(x & 0x07))) != 0) return(1);
 return(0);
}

/* 2 bit planes */
class Raster1D_2Bit:virtual public Raster1DAbstract
	{
public: virtual void Allocate1D(int NewSize1D)	{Data1D=malloc((NewSize1D+3)/4);};

	virtual int GetPlanes(void) const {return(2);};

	virtual unsigned long GetValue1D(int x) const;
	virtual void SetValue1D(int x, unsigned long NewValue);
	};

void Raster1D_2Bit::SetValue1D(int x, unsigned long NewValue)
{
BYTE *b=(BYTE *)Data1D;
BYTE v;

 v=NewValue;
 if(NewValue>3) v=3;
 b+= x >> 2;
 switch(x & 3)
   {
   case 0:v=v << 6; *b=*b & 0x3F;	break;
   case 1:v=v << 4; *b=*b & 0xCF;	break;
   case	2:v=v << 2; *b=*b & 0xF3;	break;
   case	3:*b=*b & 0xFC;
   }
 *b=*b | v;
}

void Raster1DAbstract::Get2Bit(void *Buffer2Bit)
{
 Raster1D_2Bit R1;
 R1.Data1D=Buffer2Bit;R1.Size1D=Size1D;
 Get(R1);
 R1.Data1D=NULL;
}


unsigned long Raster1D_2Bit::GetValue1D(int x) const
{
BYTE *b=(BYTE *)Data1D;

 b+= x >> 2;
 switch(x & 3)
   {
   case 0:return( (*b >> 6)&3 );
   case 1:return( (*b >> 4)&3 );
   case	2:return( (*b >> 2)&3 );
   case	3:return( *b & 3 );
   }
return(0);
}


/* 4 bit planes */
class Raster1D_4Bit:virtual public Raster1DAbstract
	{
public: virtual void Allocate1D(int NewSize1D)	{Data1D=malloc((NewSize1D+1)/2);};

	virtual int GetPlanes(void) const {return(4);};

	virtual unsigned long GetValue1D(int x) const;
	virtual void SetValue1D(int x, unsigned long NewValue);
	};


void Raster1DAbstract::Get4Bit(void *Buffer4Bit)
{
 Raster1D_4Bit R1;
 R1.Data1D=Buffer4Bit;R1.Size1D=Size1D;
 Get(R1);
 R1.Data1D=NULL;
}



void Raster1D_4Bit::SetValue1D(int x, unsigned long NewValue)
{
BYTE *b=(BYTE *)Data1D;

 if(x>=Size1D) return;
 b+= x >> 1;
 if (x & 1) *b=(*b & 0xF0) | (NewValue & 0x0F);
       else *b=(*b & 0x0F) | ((NewValue << 4) & 0xF0);
}


unsigned long Raster1D_4Bit::GetValue1D(int x) const
{
BYTE *b=(BYTE *)Data1D;

 b+= x >> 1;
 if(x & 1) return(*b & 0x0F);
      else return(*b >> 4);
}


/* 8 bit planes */
class Raster1D_8Bit:virtual public Raster1DAbstract
	{
public:	Raster1D_8Bit(int InitSize1D) {if((Data1D=malloc(InitSize1D))!=NULL) Size1D=InitSize1D;}
	Raster1D_8Bit(void)	{};

	virtual void Allocate1D(int NewSize1D)	{Data1D=malloc(NewSize1D);};

	virtual int GetPlanes(void) const {return(8);};

	virtual unsigned long GetValue1D(int x) const {if(x<Size1D) return(((BYTE *)Data1D)[x]);return(0);};
	virtual void SetValue1D(int x, unsigned long NewValue) {if(Size1D<x) return;((BYTE *)Data1D)[x]=NewValue;};
	};

void Raster1DAbstract::Join8Bit(void *Buffer8Bit,char plane8)
{
 Raster1D_8Bit R8;
 int i,maxi;
 long mask;

 R8.Data1D=Buffer8Bit;R8.Size1D=Size1D;
 maxi=Size1D>R8.Size1D?Size1D:R8.Size1D;
 mask=~(0xFFl<<plane8);

 //printf("%d %lX %d;",plane8,mask,maxi);

 for(i=0;i<maxi;i++)
	{
	SetValue1D(i,(GetValue1D(i) & mask) | (R8.GetValue1D(i) << plane8));
	}
 R8.Data1D=NULL;
}

/* 16 bit planes */
class Raster1D_16Bit:virtual public Raster1DAbstract
	{
public: Raster1D_16Bit(int InitSize1D) {if((Data1D=malloc(2*InitSize1D))!=NULL) Size1D=InitSize1D;}
	Raster1D_16Bit(void)	{};

	virtual void Allocate1D(int NewSize1D)	{Data1D=malloc(2*NewSize1D);};

	virtual int GetPlanes(void) const {return(16);};

	virtual unsigned long GetValue1D(int x) const {if(Size1D<=x) return(0); return(((WORD *)Data1D)[x]);};
	virtual void SetValue1D(int x, unsigned long NewValue) {if(Size1D<=x) return; ((WORD *)Data1D)[x]=NewValue;};

	virtual void Get8Bit(void *Buffer8Bit);
	virtual void Set8Bit(void *Buffer8Bit);
	};


void Raster1D_16Bit::Get8Bit(void *Buffer8Bit)
{
BYTE *N_Buffer8Bit = (BYTE *)Buffer8Bit;
WORD *InternalWData = (WORD *)Data1D;
int i;
for(i=0;i<Size1D;i++)
	{
	*N_Buffer8Bit++ = (*InternalWData++)>>8;
	}
}


void Raster1D_16Bit::Set8Bit(void *Buffer8Bit)
{
BYTE *N_Buffer8Bit = (BYTE *)Buffer8Bit;
WORD *InternalWData = (WORD *)Data1D;
int i;
for(i=0;i<Size1D;i++)
	{
	*InternalWData++ = (*N_Buffer8Bit++)<<8;
	}
}


/* 24 bit planes */
class Raster1D_24Bit:virtual public Raster1DAbstract
	{
public: virtual void Allocate1D(int NewSize1D)	{Data1D=malloc(3*NewSize1D);};

	virtual int GetPlanes(void) const {return(24);};

	virtual unsigned long GetValue1D(int x) const;
	virtual void SetValue1D(int x, unsigned long NewValue);

	virtual void Get24Bit(void *Buffer24Bit);

        virtual void Peel8Bit(void *Buffer8Bit,char plane8);
	};


void Raster1D_24Bit::SetValue1D(int x, unsigned long NewValue)
{
BYTE *BuffPos;

 if(x>=Size1D) return;
 BuffPos=((BYTE *)Data1D)+(3*x);
 *BuffPos++=NewValue & 0xFF;
 NewValue >>= 8;
 *BuffPos++=NewValue & 0xFF;
 NewValue >>= 8;
 *BuffPos=NewValue & 0xFF;
}


unsigned long Raster1D_24Bit::GetValue1D(int x) const
{
BYTE *b=(BYTE *)Data1D;

 if(x>=Size1D || Data1D==NULL) return(0);
 b+= 3*x;
 return((long)*b + (long)256*b[1] + (long)65536*b[2]);
}


void Raster1D_24Bit::Peel8Bit(void *Buffer8Bit,char plane8)
{
int i;
BYTE *b=(BYTE *)Data1D;
  b+=plane8;
  for(i=0;i<Size1D;i++)
    {
    *(BYTE *)Buffer8Bit=*b;
#ifdef NO_LVAL_CAST
    Buffer8Bit=(BYTE *)Buffer8Bit + 1;
#else
    ((BYTE *)Buffer8Bit)++;
#endif
    b+=3;
    }
}


void Raster1D_24Bit::Get24Bit(void *Buffer24Bit)
{
  memcpy(Buffer24Bit,Data1D,3*Size1D);
}


/* 32 bit planes */
class Raster1D_32Bit:virtual public Raster1DAbstract
	{
public: virtual void Allocate1D(int NewSize1D)	{Data1D=malloc(4*NewSize1D);};

	virtual int GetPlanes(void) const {return(32);};

	virtual unsigned long GetValue1D(int x) const {if(x<Size1D) return(((DWORD *)Data1D)[x]); return(0);};
	virtual void SetValue1D(int x, unsigned long NewValue) {if(x<Size1D) ((DWORD *)Data1D)[x]=NewValue; return;};
	};



Raster1DAbstract *CreateRaster1D(int SizeX, int Planes)
{
Raster1DAbstract *Raster=NULL;

switch(Planes)
   {
   case  0:return(NULL);
   case  1:Raster=new Raster1D_1Bit;   break;
   case  2:Raster=new Raster1D_2Bit;   break;
   case  4:Raster=new Raster1D_4Bit;   break;
   case  8:Raster=new Raster1D_8Bit;   break;
   case 16:Raster=new Raster1D_16Bit;  break;
   case 24:Raster=new Raster1D_24Bit;  break;
   case 32:Raster=new Raster1D_32Bit;  break;
   default:return(NULL);
   }
if(Raster)
   {
   Raster->Allocate1D(SizeX);
   if(Raster->Data1D==NULL)
	{
	delete Raster;
	return(NULL);
	}
   else Raster->Size1D=SizeX;
   }

return(Raster);
}


/* Raster 2D abstract class */

void Raster2DAbstract::Erase2DStub(void)
{
int i,SaveSize1D;
if(Data2D)
  {
  SaveSize1D=Size1D;
  for(i=0;i<Size2D;i++)
	{
	Size1D=SaveSize1D;
	Data1D=Data2D[i];
	Raster1DAbstract::Erase1D();
	}
  free(Data2D);
  Data2D=NULL;
  }
Size2D=0;
}


void Raster2DAbstract::Cleanup(void)
{
int i;

 if(Data2D==NULL) return;
 for(i=0;i<Size2D;i++)
	{	
	Data1D=Data2D[i];
	Raster1DAbstract::Cleanup();
	}
}


void Raster2DAbstract::Allocate2D(int NewSize1D, int NewSize2D)
{
int i;

if(Data2D) Erase();

Data2D=(void **)malloc(sizeof(void *)*NewSize2D);
if(!Data2D) return;		/*Not Enough memory*/
memset(Data2D,0,sizeof(void *)*NewSize2D);

Size2D=NewSize2D;
for(i=0;i<NewSize2D;i++)
	{
	Size1D=NewSize1D;
	Data1D=NULL;
	Allocate1D(Size1D);
	Data2D[i]=Data1D;
	if(Data1D==NULL)
		{
		Erase();
		return;
		}
	}
}

Raster1DAbstract *Raster2DAbstract::GetRowRaster(int Offset2D)
{
 if(Data2D==NULL || Offset2D>=Size2D) return(NULL);
 Data1D=Data2D[Offset2D];
 return(this);
}


long Raster2DAbstract::GetValue2D(int Offset1D,int Offset2D)
{
 if(Data2D==NULL || Offset2D>=Size2D) return(0);
 Data1D=Data2D[Offset2D];
 return(GetValue1D(Offset1D));
}

void Raster2DAbstract::SetValue2D(int Offset1D,int Offset2D,long x)
{
 if(Data2D==NULL || Offset2D>=Size2D) return;
 Data1D=Data2D[Offset2D];
 SetValue1D(Offset1D,x);
}

/* Specialised Raster 2D modules */
class Raster2D_1Bit:public Raster2DAbstract,public Raster1D_1Bit {};
class Raster2D_2Bit:public Raster2DAbstract,public Raster1D_2Bit {};
class Raster2D_4Bit:public Raster2DAbstract,public Raster1D_4Bit {};
class Raster2D_8Bit:public Raster2DAbstract,public Raster1D_8Bit {};
class Raster2D_16Bit:public Raster2DAbstract,public Raster1D_16Bit {};
class Raster2D_24Bit:public Raster2DAbstract,public Raster1D_24Bit {};
class Raster2D_32Bit:public Raster2DAbstract,public Raster1D_32Bit {};


Raster2DAbstract *CreateRaster2D(int SizeX, int SizeY, int Planes)
{
Raster2DAbstract *Raster=NULL;

switch(Planes)
   {
   case  0:return(NULL);
   case  1:Raster=new Raster2D_1Bit;   break;
   case  2:Raster=new Raster2D_2Bit;   break;
   case  4:Raster=new Raster2D_4Bit;   break;
   case  8:Raster=new Raster2D_8Bit;   break;
   case 16:Raster=new Raster2D_16Bit;  break;
   case 24:Raster=new Raster2D_24Bit;  break;
   case 32:Raster=new Raster2D_32Bit;  break;
   default:return(NULL);
   }
if(Raster)
   {
   Raster->Allocate2D(SizeX,SizeY);
   if(Raster->Data2D==NULL)
	{
	delete Raster;
	return(NULL);
	}
   }

return(Raster);
}


#ifdef R3D
/* Raster 3D abstract class */

void Raster3DAbstract::Erase3DStub(void)
{
int i,SaveSize2D;
if(Data3D)
  {
  SaveSize2D=Size2D;
  for(i=0;i<Size3D;i++)
	{
	Size2D=SaveSize2D;
	Data2D=Data3D[i];
	Raster2DAbstract::Erase2D();
	}
  free(Data3D);
  Data3D=NULL;
  }
Size3D=0;
}


void Raster3DAbstract::Cleanup(void)
{
int i;

 if(Data3D==NULL) return;
 for(i=0;i<Size3D;i++)
	{	
	Data2D=Data3D[i];
	Raster2DAbstract::Cleanup();
	}
}


void Raster3DAbstract::Allocate3D(int NewSize1D, int NewSize2D, int NewSize3D)
{
int i;

if(Data3D) Erase();

Data3D=(void ***)malloc(sizeof(void *)*NewSize3D);
if(!Data3D) return;		/*Not Enough memory*/
memset(Data3D,0,sizeof(void *)*NewSize3D);

Size3D=NewSize3D;
for(i=0;i<NewSize3D;i++)
	{
	Size2D=NewSize2D;
	Data2D=NULL;
	Allocate2D(NewSize1D,Size2D);
	Data3D[i]=Data2D;
	if(Data2D==NULL)
		{
		Erase();
		return;
		}
	}
}

Raster2DAbstract *Raster3DAbstract::GetRowRaster(int Offset3D)
{
 if(Data3D==NULL || Offset3D>=Size3D) return(NULL);
 Data2D=Data3D[Offset3D];
 return(this);
}


long Raster3DAbstract::GetValue3D(int Offset1D,int Offset2D,int Offset3D)
{
 if(Data3D==NULL || Offset3D>=Size3D) return(0);
 Data2D=Data3D[Offset3D];
 return(GetValue2D(Offset1D,Offset2D));
}

void Raster3DAbstract::SetValue3D(int Offset1D,int Offset2D,int Offset3D,long x)
{
 if(Data3D==NULL || Offset3D>=Size3D) return;
 Data2D=Data3D[Offset3D];
 SetValue2D(Offset1D,Offset2D,x);
}

/* Specialised Raster 3D modules */
class Raster3D_1Bit:virtual public Raster3DAbstract,virtual public Raster2D_1Bit {};
class Raster3D_2Bit:virtual public Raster3DAbstract,virtual public Raster2D_2Bit {};
class Raster3D_4Bit:virtual public Raster3DAbstract,virtual public Raster2D_4Bit {};
class Raster3D_8Bit:virtual public Raster3DAbstract,virtual public Raster2D_8Bit {};
class Raster3D_16Bit:virtual public Raster3DAbstract,virtual public Raster2D_16Bit {};
class Raster3D_24Bit:virtual public Raster3DAbstract,virtual public Raster2D_24Bit {};
class Raster3D_32Bit:virtual public Raster3DAbstract,virtual public Raster2D_32Bit {};


Raster3DAbstract *CreateRaster3D(int SizeX, int SizeY, int SizeZ, int Planes)
{
Raster3DAbstract *Raster=NULL;

switch(Planes)
   {
   case  0:return(NULL);
   case  1:Raster=new Raster3D_1Bit;   break;
   case  2:Raster=new Raster3D_2Bit;   break;
   case  4:Raster=new Raster3D_4Bit;   break;
   case  8:Raster=new Raster3D_8Bit;   break;
   case 16:Raster=new Raster3D_16Bit;  break;
   case 24:Raster=new Raster3D_24Bit;  break;
   case 32:Raster=new Raster3D_32Bit;  break;
   default:return(NULL);
   }
if(Raster)
   {
   Raster->Allocate3D(SizeX,SizeY,SizeZ);
   if(Raster->Data3D==NULL)
	{
	delete Raster;
	return(NULL);
	}
   }

return(Raster);
}

#endif

/* ---------- Global class Image --------- */
void Image::Create(int SizeX, int SizeY, int Planes)
{
if(Raster)
	{
	if(Raster->Size1D==SizeX && Raster->Size2D==SizeY && Raster->Planes==Planes && Raster->UsageCount==1) return;
	Erase();
	}

if( (Raster=CreateRaster2D(SizeX,SizeY,Planes))!=NULL)
	Raster->UsageCount++;
}

Image::Image(const Image & I)
{
Image *PImg,*CurrImg;
 if((Raster=I.Raster)!=NULL) Raster->UsageCount++;
 if((Palette=I.Palette)!=NULL) Palette->UsageCount++;
 x=I.x;y=I.y;dx=I.dx;dy=I.dy;RotAngle=I.RotAngle;
 Next=NULL;
 CurrImg=I.Next;
 Comment=NULL;
 if(I.Comment!=NULL) Comment=strdup(I.Comment);
 PImg=this;
 while(CurrImg!=NULL)	//Multiple images
	{
	PImg->Next=new Image;
	if(PImg->Next==NULL) return;	/*Allocation error*/
	PImg=PImg->Next;
	if((PImg->Raster=CurrImg->Raster)!=NULL) CurrImg->Raster->UsageCount++;
	if((PImg->Palette=CurrImg->Palette)!=NULL) CurrImg->Palette->UsageCount++;
	PImg->x=CurrImg->x;PImg->y=CurrImg->y;
	PImg->dx=CurrImg->dx;PImg->dy=CurrImg->dy;
	PImg->RotAngle=CurrImg->RotAngle;
	if(CurrImg->Comment!=NULL) PImg->Comment=strdup(CurrImg->Comment);
	CurrImg=CurrImg->Next;
	}
}

Raster2DAbstract *Image::operator=(Raster2DAbstract *NewRaster)
{
 x=y=dx=dy=0;
 Comment=NULL;
 if(Raster==NewRaster) return(Raster);		//Identical assignment
 if(Raster!=NULL) Erase();
 if( (Raster=NewRaster)!=NULL ) Raster->UsageCount++;
 return(Raster);
}

Image &Image::operator=(const Image &I)
{
Image *PImg;
const Image *CurrImg;
 Erase();

 PImg=this;
 CurrImg=&I;
 goto CopyImg;

 while(CurrImg!=NULL)	//Multiple images
	{
	PImg->Next=new Image;
	if(PImg->Next==NULL) return(*this);
	PImg=PImg->Next;

CopyImg:if((PImg->Raster=CurrImg->Raster)!=NULL)
		CurrImg->Raster->UsageCount++;
	if((PImg->Palette=CurrImg->Palette)!=NULL)
		CurrImg->Palette->UsageCount++;
	PImg->x=CurrImg->x;PImg->y=CurrImg->y;
	PImg->dx=CurrImg->dx;PImg->dy=CurrImg->dy;
	PImg->RotAngle=CurrImg->RotAngle;
	if(CurrImg->Comment!=NULL) PImg->Comment=strdup(CurrImg->Comment);
	CurrImg=CurrImg->Next;
	}
 return(*this);
}

void Image::Erase(void)
{
Image *I,*OldI;

 I=this;
 while(I!=NULL)
   {
   if(I->Raster!=NULL)
      {
      if(--I->Raster->UsageCount<=0) delete I->Raster;
      I->Raster=NULL;
      }
   if(I->Palette!=NULL)
      {
      if(--I->Palette->UsageCount<=0) delete I->Palette;
      I->Palette=NULL;
      }
   if(I->Comment!=NULL)
      {
      free(I->Comment);
      I->Comment=NULL;
      }
   OldI=I;
   I=I->Next;
   OldI->Next=NULL;
   if(OldI!=this) delete(OldI);
   }
}


/* 0-none, 1-gray, 2-palette, 3-true color */
IMAGE_TYPE Image::ImageType(void) const
{
  if(Raster==NULL) return ImageNone;
  if(Palette!=NULL)
    {
    if(GrayPalette(Palette,Raster->GetPlanes())) return ImageGray;
    return ImagePalette;
    }
  if(Raster->GetPlanes()==24) return ImageTrueColor;
  return ImageGray;
}


/* ---------------------------- */


void LoadImage(Image & x)
{
 x.SetPixel(5,5,100);

// x.Erase();
// x.Create(10,10,8);

}


/*This procedure provides vertical flipping of the raster*/
void Flip2D(Raster2DAbstract *r2D)
{
void *ptr,**P1,**P2;

if(r2D==NULL) return;
if(r2D->Data2D==NULL) return;
P1=r2D->Data2D;
P2=r2D->Data2D+(r2D->Size2D-1);
while(P2>P1)
   {
   ptr=*P1;
   *P1=*P2;
   *P2=ptr;
   P1++; P2--;
   }
}

/*This procedure provides horizontal flipping of the raster*/
void Flip1D(Raster2DAbstract *r2D)
{
WORD *Buff,*B1,*B2,tmp;
unsigned y;

if(r2D==NULL) return;
if(r2D->Data2D==NULL) return;

  Buff=(WORD *)malloc(r2D->Size1D*sizeof(WORD));
  for(y=0;y<r2D->Size2D;y++)
      {
      r2D->GetRowRaster(y)->Get16Bit(Buff);

      B1=Buff;
      B2=Buff+r2D->Size1D-1;

      while(B1<B2)
	 {
	 tmp=*B1;
	 *B1=*B2;
	 *B2=tmp;
	 B1++; B2--;
	 }
      r2D->GetRowRaster(y)->Set16Bit(Buff);
      }
  free(Buff);
}


class Palette_8bit:virtual public APalette,virtual public Raster1D_8Bit
	{
public:	Palette_8bit(int Indices):Raster1D_8Bit(3*Indices) {};
	Palette_8bit(void) {};
	virtual void Get(unsigned index, RGBQuad *RGB);
	};
	
class Palette_16bit:virtual public APalette,virtual public Raster1D_16Bit
	{
public:	Palette_16bit(int Indices):Raster1D_16Bit(3*Indices) {};
	Palette_16bit(void) {};
	};

void Palette_8bit::Get(unsigned index, RGBQuad *RGB)
{
BYTE *ptrb;
 index*=3;
 if(RGB==NULL || Data1D==NULL || Size1D-3<index) return;
 ptrb=((BYTE *)Data1D)+index;
 RGB->R=*ptrb++;
 RGB->G=*ptrb++;
 RGB->B=*ptrb;
}


void APalette::Get(unsigned index,RGBQuad *RGB)
{
 if(RGB==NULL) return;
 RGB->R=R(index);
 RGB->G=G(index);
 RGB->B=B(index);
}


APalette *BuildPalette(int Indices, char type)
{
APalette *pal=NULL;

 switch(type)
	{
	case 0:
	case 8: pal=new Palette_8bit(Indices);	break;
	case 16:pal=new Palette_16bit(Indices); break;
	}
 if(pal)
   if(pal->Size1D<Indices)
	{
	delete pal;
	return(NULL);
	}
   	
return(pal);
}


int GrayPalette(Raster1DAbstract *Palette,int Planes) /*Is this Gray scale palette?*/
{
unsigned i,j;
unsigned PalItems;

 if(Palette==NULL) return(1);
 if(Palette->Data1D==NULL ||
    Palette->Size1D==0) return(1);	//consider invalid palette to be gray

 PalItems = (1<<Planes);
 if(Planes!=0)
   if(3*PalItems > Palette->Size1D)
	{
	//PalItems=Palette->Size1D/3;
	return(0);	//insufficient palette size, ?? non gray ??
	}

 for(i=0;i<PalItems;i++)
	{
	j=i*255 / (PalItems-1);
	if( (Palette->GetValue1D(3*i)!=j) ||
	    (Palette->GetValue1D(3*i+1)!=j) ||
	    (Palette->GetValue1D(3*i+2)!=j) ) return(0);
	}
 return(1);
}


int FillGray(Raster1DAbstract *Palette,int Planes) /*Make this palette gray*/
{
unsigned i,j;

 if(Palette==NULL) return(1);
 if(Planes!=0)
   if(3*(1<<Planes) != Palette->Size1D) return(0);

 if(Palette->Data1D==NULL) return(0);
 for(i=0;i<Palette->Size1D/3-1;i++)
	{
	j=i*255 / (Palette->Size1D/3-1);
	Palette->SetValue1D(3*i,j);
	Palette->SetValue1D(3*i+1,j);
	Palette->SetValue1D(3*i+2,j);
	}
 return(1);
}
